package com.sksamuel.scoot.rest.client

import javax.ws.rs.core.MediaType
import com.sun.jersey.api.client.{Client, ClientResponse, WebResource}
import com.sksamuel.scoot.rest._
import com.sun.jersey.api.client.config.DefaultClientConfig
import com.fasterxml.jackson.jaxrs.json.JacksonJsonProvider
import com.fasterxml.jackson.databind.ObjectMapper

/** @author Stephen Samuel */
case class Request(mapper: ObjectMapper,
                   path: String,
                   headers: Map[String, String],
                   credentials: Credentials,
                   queryParams: Map[String, String],
                   formParams: Map[String, String],
                   method: Method,
                   body: Option[Any],
                   bodyType: Option[String],
                   acceptTypes: Seq[String]) extends Logging {

    def authenticate(cred: Credentials) = throw new UnsupportedOperationException

    def header(key: String, value: String) = copy(headers = headers + ((key, value)))
    def userAgent(ua: String) = copy(headers = headers + ("User-Agent" -> ua))
    def referer(ua: String) = copy(headers = headers + ("Referer" -> ua))

    def accept(types: String*): Request = copy(acceptTypes = types)
    implicit def mediaType2String(mediaType: MediaType): String = mediaType.toString

    def cookie(name: String, value: String) = copy(headers = headers + (name -> value))

    def get: Response = copy(method = Get).sendReceive
    def delete: Response = copy(method = Delete).sendReceive

    def post: Response = copy(method = Post).sendReceive
    def post(body: Any, mediaType: MediaType): Response = post(body, mediaType.toString)
    def post(body: Any, mediaType: String): Response = {
        copy(body = Some(body), bodyType = Some(mediaType), method = Post).sendReceive
    }

    def put: Response = copy(method = Put).sendReceive
    def put(body: Any, mediaType: MediaType): Response = put(body, mediaType.toString)
    def put(body: Any, mediaType: String): Response = {
        copy(body = Some(body), bodyType = Some(mediaType), method = Put).sendReceive
    }

    def queryParam(name: String, value: Any) = copy(queryParams = queryParams + (name -> value.toString))
    def formParam(name: String, value: Any) = copy(formParams = formParams + (name -> value.toString))

    private val cc = new DefaultClientConfig()
    cc.getSingletons.add(new JacksonJsonProvider(mapper))
    private val client = Client.create(cc)

    private def sendReceive: Response = {

        val resource = client.resource(path)
        val f: (WebResource, (String, String)) => WebResource = (t, p) => t.queryParam(p._1, p._2)
        val resourceWithParams = queryParams.foldLeft(resource)(f)

        val builder = resourceWithParams.getRequestBuilder
        acceptTypes.foreach(t => builder.accept(t))
        headers.foreach(header => {
            builder.header(header._1, header._2)
        })
        body.foreach(arg => builder.entity(arg, bodyType.getOrElse(MediaType.APPLICATION_OCTET_STREAM)))

        val resp = builder.method(method.name, classOf[ClientResponse])
        logger.debug("Response [status={} headers={}, body={}]", resp.getStatus, resp.getHeaders)
        new Response(resp)
    }
}