package com.sksamuel.scoot.rest.client2

import java.io.{ByteArrayOutputStream, InputStream}
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.scala.DefaultScalaModule
import javax.ws.rs.core.MediaType
import javax.xml.bind.JAXBContext
import org.apache.commons.io.IOUtils

/** @author Stephen Samuel */
trait EntityMarshaller {
    def unmarshall[T: Manifest](input: InputStream): T
    def marshall(t: Any): Array[Byte]
    def isDefinedFor(mediaType: String): Boolean
}

object EntityMarshaller {
    val marshallers = Seq(new JacksonJsonMarshaller, new JaxbXmlMarshaller, new TextEntityConvertor)
    def marshaller(mediaType: String): Option[EntityMarshaller] = marshallers.find(_.isDefinedFor(mediaType))
}

class JacksonJsonMarshaller extends EntityMarshaller {

    val mapper = new ObjectMapper
    mapper.registerModule(DefaultScalaModule)

    def isDefinedFor(mediaType: String) = mediaType == MediaType.APPLICATION_JSON

    def unmarshall[T: Manifest](input: InputStream): T = {
        val qqq = mapper.getTypeFactory
          .constructParametricType(manifest.runtimeClass, manifest.typeArguments.map(_.runtimeClass): _*)
        mapper.readValue(input, qqq)
    }

    def marshall(t: Any): Array[Byte] = mapper.writeValueAsBytes(t)
}

class JaxbXmlMarshaller extends EntityMarshaller {
    def isDefinedFor(mediaType: String) = mediaType == MediaType.TEXT_XML
    def unmarshall[T: Manifest](input: InputStream): T = JAXBContext
      .newInstance(manifest.runtimeClass)
      .createUnmarshaller()
      .unmarshal(input).asInstanceOf[T]

    def marshall(t: Any): Array[Byte] = {
        val out = new ByteArrayOutputStream
        JAXBContext.newInstance(t.getClass).createMarshaller().marshal(t, out)
        out.toByteArray
    }
}

class TextEntityConvertor extends EntityMarshaller {
    def isDefinedFor(mediaType: String) = mediaType == MediaType.TEXT_PLAIN
    def unmarshall[T: Manifest](input: InputStream): T = IOUtils.toString(input, "utf-8").asInstanceOf[T]
    def marshall(t: Any): Array[Byte] = t.toString.getBytes("utf-8")
}