package com.sksamuel.scoot.rest.client2

import com.sksamuel.scoot.rest.client.{NoCredentials, Credentials}
import com.sksamuel.scoot.rest._
import javax.ws.rs.core.MediaType
import scala.concurrent._
import org.apache.http.impl.client.DefaultHttpClient
import org.apache.http.client.methods.{HttpDelete, HttpPut, HttpPost, HttpGet}
import org.apache.http.{Header, HttpEntity, HttpResponse}
import org.apache.http.entity.ByteArrayEntity
import org.apache.http.client.utils.URIBuilder
import scala.util.{DynamicVariable, Success, Failure, Try}
import ExecutionContext.Implicits._
import scala.UnsupportedOperationException
import org.apache.http.client.HttpClient
import org.apache.http.impl.conn.{SchemeRegistryFactory, PoolingClientConnectionManager}
import scala.concurrent.duration._
import scala.collection.mutable.ListBuffer
import scala.collection.mutable

/** @author Stephen Samuel */
class ScootClient(client: HttpClient, marshallers: Seq[EntityMarshaller]) {

    val context = new DynamicVariable[Option[mutable.Buffer[ScootRequest]]](None)

    def get(url: String): ScootRequest = {
        _request(url, Get)
    }

    def post(url: String): ScootRequest =
        _request(url, Post)

    def put(url: String): ScootRequest =
        _request(url, Put)

    def delete(url: String): ScootRequest =
        _request(url, Delete)

    def _request(url: String, method: Method): ScootRequest = {
        val req = new ScootRequest(url = url, method = method)(client)
        context.value.foreach(_.append(req))
        req
    }

    // each request in this block will be executed synchronously, blocking for each one in return
    def inSequence(f: => Seq[ScootRequest]): Seq[ScootResponse] = {
        val buffer = new ListBuffer[ScootRequest]
        context.withValue(Some(buffer))(f)
        buffer.map(_.execute)
    }

    // each request in this block will be executed asynchronously, returning the set of all futures
    def inAnyOrder(f: => Seq[ScootRequest]): Seq[Future[ScootResponse]] = {
        val buffer = new ListBuffer[ScootRequest]
        context.withValue(Some(buffer))(f)
        buffer.map(_.async)
    }

    // a blocking operation that will block on a sequence of asynchronous operations. That is, all the individual
    // requests are non blocking but the overall set of requests will block until all are completed.
    def batch(f: => Unit): Seq[Future[ScootResponse]] = batch(f, 10 seconds)

    // a blocking operation that will block on a sequence of asynchronous operations. That is, all the individual
    // requests are non blocking but the overall set of requests will block until all are completed.
    def batch(f: => Unit, duration: Duration): Seq[Future[ScootResponse]] = {
        val buffer = new ListBuffer[ScootRequest]
        context.withValue(Some(buffer))(f)
        val futures = buffer.map(_.async)
        futures.foreach(Await.ready(_, duration))
        futures
    }
}

object ScootClient {
    // creates a new client with the default set of marshallers
    // and using a pre-configured Apache HTTP client
    def apply(): ScootClient = apply(EntityMarshaller.marshallers)
    def apply(marshallers: Seq[EntityMarshaller]): ScootClient = {
        val connManager = new PoolingClientConnectionManager(SchemeRegistryFactory.createDefault())
        connManager.setMaxTotal(200)
        connManager.setDefaultMaxPerRoute(200)
        val client = new DefaultHttpClient(connManager)
        new ScootClient(client, marshallers)
    }
}

case class ScootRequest(url: String,
                        headers: Map[String, String] = Map.empty,
                        credentials: Credentials = NoCredentials,
                        queryParams: Map[String, String] = Map.empty,
                        formParams: Map[String, String] = Map.empty,
                        method: Method,
                        entity: Option[Any] = None,
                        mediaType: String = MediaType.WILDCARD,
                        acceptTypes: Seq[String] = Nil)(implicit client: HttpClient) extends Logging {

    implicit def mediaType2String(mediaType: MediaType): String = mediaType.toString

    def authenticate(cred: Credentials) = throw new UnsupportedOperationException

    def header(kv: (String, Any)): ScootRequest = copy(headers = headers + ((kv._1, kv._2.toString)))
    def header(key: String, value: Any): ScootRequest = header((key, value))

    import Headers._

    def userAgent(ua: String): ScootRequest = copy(headers = headers + (UserAgent -> ua))
    def referer(ua: String): ScootRequest = copy(headers = headers + (Referer -> ua))

    def accept(types: String*): ScootRequest = copy(acceptTypes = types)

    def queryParam(name: String, value: Any): ScootRequest = copy(queryParams = queryParams + (name -> value
      .toString))
    def formParam(name: String, value: Any): ScootRequest = copy(formParams = formParams + (name -> value.toString))
    def cookie(name: String, value: String): ScootRequest = copy(headers = headers + (name -> value))

    // sets an entity onto the request body
    def entity(entity: Any): ScootRequest = copy(entity = Option(entity))
    def entity(entity: Any, mediaType: String): ScootRequest = copy(entity = Option(entity), mediaType = mediaType)
    def entity(entity: Any, mediaType: MediaType): ScootRequest = copy(entity = Option(entity),
        mediaType = mediaType)

    // execute this request asyncronously
    def async: Future[ScootResponse] = {
        future {
            execute
        }
    }

    def execute: ScootResponse = {

        val uri = new URIBuilder(url)
        queryParams.foreach(param => uri.addParameter(param._1, param._2))

        val r = method match {

            case Get => new HttpGet(uri.build)
            case Delete => new HttpDelete(uri.build)

            case Post =>
                val post = new HttpPost(uri.build)
                if (entity.isDefined) {
                    val marshaller = EntityMarshaller
                      .marshaller(mediaType)
                      .getOrElse(throw new UnsupportedOperationException(mediaType + " is not supported"))
                    post.setEntity(new ByteArrayEntity(marshaller.marshall(entity)))
                }
                post

            case Put =>
                val put = new HttpPut(uri.build)
                if (entity.isDefined) {
                    val marshaller = EntityMarshaller
                      .marshaller(mediaType)
                      .getOrElse(throw new UnsupportedOperationException(mediaType + " is not supported"))
                    put.setEntity(new ByteArrayEntity(marshaller.marshall(entity)))
                }
                put

        }
        logger.debug(s"Executing underlying apache request $r")
        val resp = client.execute(r)
        ScootResponse(resp)
    }
}

class ScootResponse(val status: Int, _entity: HttpEntity, headers: Seq[Header]) {

    implicit def request2response(request: ScootRequest): ScootResponse = request.execute

    import Headers._

    def header(name: String): Option[String] = headers.find(_.getName == name).map(_.getValue)
    def cookie(name: String): Option[Cookie] = None
    def length = header(ContentLength)
    def contentType = header(ContentType)
    def close(): Unit = Option(_entity).foreach(_.getContent.close())

    private def _safeInvoke[T: Manifest]: Try[T] = {
        try {
            val marshaller = EntityMarshaller
              .marshaller(contentType.getOrElse(MediaType.WILDCARD))
              .getOrElse(throw new UnsupportedOperationException(contentType + " is not supported"))
            val t = marshaller.unmarshall(_entity.getContent)
            Success(t)
        } catch {
            case e: Exception => Failure(e)
        } finally {
            close()
        }
    }

    def entity[T: Manifest]: Try[T] = _safeInvoke
    override def toString = s"ScootResponse [status=$status headers=$headers]"
}

object ScootResponse {
    def apply(resp: HttpResponse): ScootResponse = {
        val status = resp.getStatusLine.getStatusCode
        val entity = resp.getEntity
        val headers = resp.getAllHeaders
        new ScootResponse(status, entity, headers)
    }
}

case class Cookie(name: String, value: String, path: String, maxAge: Int, domain: String)